import math
from typing import List
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from ._nnls import (
    nnls,
)


# pyplot.legend() -> gca().legend()
# axes.legend() -> legend._parse_legend_args()
# legend._parse_legend_args() -> labels!=None and handles==None:
def get_dataline_handles(
    ax: mpl.axes.Axes,
):
    """Returns the handles that correspond to datalines

    Args:
        ax (Axes): the axis of the graph to look in
    Returns:
        list[handle]: list of handles that correspond to datalines
    """
    return [
        h for h in mpl.legend._get_legend_handles([ax]) if type(h) == mpl.lines.Line2D and not np.all(np.isnan(h.get_ydata()))
    ]


def is_line_plot(ax: mpl.axes.Axes, labels: List[str] = None) -> bool:
    """Determines whether the plot is a pure line plot or if it contains other types of data visualizations.

    Args:
        ax (mpl.axes.Axes): Axes object of the graph
        labels (list[str], optional): list of labels to use. Defaults to None.
    Returns:
        bool: True if the plot is a pure lineplot, False otherwise
    """
    handles = get_dataline_handles(ax)
    if len(handles) == 0:
        return False  # No lines in plot
    if labels is not None and len(handles) != len(labels):
        # Some plots (e.g. violin/bar plots) also use Line2D object for some data (e.g. errorlines))
        return False
    # Filter scatterplots
    linestyles = [h._linestyle for h in handles]
    return "None" not in linestyles


def get_linelegend_ypositions(
    ax: mpl.axes.Axes,
    handles: List[mpl.artist.Artist],
    labels: List[str] = None,
):
    """Calculates the positions of the legendentries.
    Ensures that there is no vertical overlap between entires.
    https://github.com/nschloe/dufte?src/dufte/main.py

    Args:
        ax (mpl.axes.Axes): Axes object of the graph
        handles (list[mpl.lines.Line2D]): list of handles representing the datalines we are making a legend for.
        labels (list[str], optional): list of labels to use.
            Defaults to None.
    Returns:
        list[(float, float)]: a list of (x,y) coordinates for the legend entries
    """
    fontsize_inches = mpl.rcParams["font.size"] / 72
    fig_height_inches = plt.gcf().get_size_inches()[1]
    yaxis_height_inches = (ax.get_position().y1 - ax.get_position().y0) * fig_height_inches
    if ax.get_yscale() == "log":
        yaxis_range = math.log10(ax.get_ylim()[1]) - math.log10(ax.get_ylim()[0])
    else:
        yaxis_range = ax.get_ylim()[1] - ax.get_ylim()[0]
    spacing_between_lines = 1.1
    line_height = fontsize_inches / yaxis_height_inches * yaxis_range * spacing_between_lines

    # Find position heights and adjust so that there is no overlap
    last_y = [h.get_ydata()[~np.isnan(h.get_ydata())][-1] for h in handles]
    if ax.get_yscale() == "log":
        last_y = [math.log10(y) for y in last_y]
    idx = np.argsort(last_y)
    targets = np.sort(last_y)
    labels = np.array(labels)[idx]
    n = len(targets)
    lines_away_from_first_label = np.arange(
        n,
        dtype=float,
    )
    for i in range(
        1,
        len(labels),
    ):
        distance_to_prev_label = (2 + f"{labels[i-1]}_{labels[i]}".count("\n")) / 2
        lines_away_from_first_label[i] = lines_away_from_first_label[i - 1] + distance_to_prev_label

    y0_min = targets[0] - lines_away_from_first_label[-1] * line_height
    A = np.tril(np.ones([n, n]))
    b = targets - (y0_min + lines_away_from_first_label * line_height)

    out = nnls(A, b)
    sol = np.cumsum(out) + y0_min + lines_away_from_first_label * line_height
    idx2 = np.argsort(idx)
    last_y = sol[idx2]
    if ax.get_yscale() == "log":
        last_y = [10**y for y in last_y]

    # Add x-coordinates to get position next to last datapoint
    last_x = [h.get_xdata()[~np.isnan(h.get_xdata())][-1] for h in handles]
    xaxis_range = ax.get_xlim()[1] - ax.get_xlim()[0]
    targets = [
        (
            x + xaxis_range * 0.03,
            y,
        )
        for x, y in zip(
            last_x,
            last_y,
        )
    ]

    return targets



import numpy as np


# src: https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.nnls.html
# src: https://github.com/nschloe/dufte/blob/main/src/dufte/optimize.py
def nnls(
    A,
    b,
    eps: float = 1.0e-10,
    max_steps: int = 100,
):
    A = np.asarray(A)
    b = np.asarray(b)

    AtA = A.T @ A
    Atb = A.T @ b

    m, n = A.shape
    assert m == b.shape[0]
    mask = np.zeros(
        n,
        dtype=bool,
    )
    x = np.zeros(n)
    w = Atb
    s = np.zeros(n)
    k = 0
    while sum(mask) != n and max(w) > eps:
        if k >= max_steps:
            break
        mask[np.argmax(w)] = True

        s[mask] = np.linalg.lstsq(
            AtA[mask][
                :,
                mask,
            ],
            Atb[mask],
            rcond=None,
        )[0]
        s[~mask] = 0.0

        while np.min(s[mask]) <= 0:
            alpha = np.min(x[mask] / (x[mask] - s[mask]))
            x += alpha * (s - x)
            mask[np.abs(x) < eps] = False

            s[mask] = np.linalg.lstsq(
                AtA[mask][
                    :,
                    mask,
                ],
                Atb[mask],
                rcond=None,
            )[0]
            s[~mask] = 0.0

        x = s.copy()
        w = Atb - AtA @ x

        k += 1

    return x



from .style import (
    style,
)
from .functions import (
    legend,
    ylabel,
    show_bar_values,
)
from ._colors import (
    colors,
)

__all__ = [
    "style",
    "colors",
    "legend",
    "ylabel",
    "show_bar_values",
]



from pathlib import (
    Path,
)

_fonts_dir = Path(__file__).parent.parent.joinpath("fonts")
fonts = {
    "Nunito": Path(
        _fonts_dir,
        "Nunito_Sans",
        "NunitoSans-Regular.ttf",
    ),
    "Arvo": Path(
        _fonts_dir,
        "Arvo",
        "Arvo-Regular.ttf",
    ),
    "Inconsolata": Path(
        _fonts_dir,
        "Inconsolata",
        "Inconsolata-Regular.ttf",
    ),
}



"""One of the main files of rootsstyle. This file defines the added functionalities:
- rootsstyle.legend()
- rootsstyle.ylabel()
"""
from ._fonts import (
    fonts,
)
from ._colors import (
    layout_colors,
    palettes,
)
from ._utils import (
    is_line_plot,
    get_dataline_handles,
    get_linelegend_ypositions,
)
from typing import List
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.font_manager as fm

# Register fonts in global FontManager
font_entries = [
    fm.FontEntry(
        fname=path,
        name=name,
    )
    for name, path in fonts.items()
]
fm.fontManager.ttflist.extend(font_entries)
# Register colormaps in global ColorMap manager
[
    mpl.cm.register_cmap(
        cmap_name,
        cmap,
    )
    for cmap_name, cmap in palettes.items()
]


# Inspiration: https://github.com/nschloe/dufte
# Inspiration: https://github.com/matplotlib/matplotlib/blob/main/lib/matplotlib/legend.py
def _legend_line(ax: mpl.axes.Axes, labels: List[str] = None):
    """Displays the legendentries next to the line in the same color as the line

    Args:
        ax (mpl.axes.Axes): Axes object of the graph
        labels (list[str], optional): list of labels to use. Defaults to None.
    Returns:
        dict: a dictionary containing entries for 'labels' and 'handles' used by the legend.
    """
    # Removing existing legend (e.g. default of seaborn)
    handles = get_dataline_handles(ax)
    colors = [h.get_color() for h in handles]
    if labels is None or len(labels) != len(handles):
        labels = [h.get_label() for h in handles]

    targets = get_linelegend_ypositions(
        ax,
        handles,
        labels,
    )

    for (label, (x, y), color,) in zip(
        labels,
        targets,
        colors,
    ):
        plt.text(
            x,
            y,
            label,
            verticalalignment="center",
            color=color,
        )
    return {
        "labels": labels,
        "handles": handles,
    }


def legend(
    handles: List[mpl.artist.Artist] = None,
    labels: List[str] = None,
    title: str = None,
):
    """Displays the legend to the left of the plot.
    For lineplots, displays each line legend next to the line.

    Args:
        handles (list[mpl.artist.Artist], optional): list of Artists (lines, patches) to be added to the legend.
            Defaults to None.
        labels (list[str], optional): A list of labels to show next to the artists.
            Defaults to None.
        title (str, optional): The legend's title.
            Defaults to None.
    """
    ax = plt.gca()
    # Check for existing title
    if ax.legend_ is not None:  # e.g. seaborn already set legend
        title = title or ax.legend_.get_title().get_text()
        handles = handles or ax.legend_.legendHandles
        labels = labels or [t.get_text() for t in ax.legend_.get_texts()]
        ax.legend_ = None

    if is_line_plot(ax, labels):
        _legend_line(
            ax,
            labels,
        )
    else:
        ax.legend(
            handles=handles,
            labels=labels,
            title=title,
            loc="center left",
            bbox_to_anchor=(
                1,
                0.5,
            ),
        )
    return {
        "labels": labels,
        "handles": handles,
    }


# Inspiration: https://github.com/nschloe/dufte
def ylabel(
    ylabel: str,
):
    """Adds a ylabel to the plot at the top of the y-axis

    Args:
        ylabel (str): the name of the label
    """
    ax = plt.gca()
    plt.ylabel(
        ylabel,
        horizontalalignment="center",
    ).set_rotation(0)
    # sits 3% above top tick
    ax.yaxis.set_label_coords(0, 1.03)


def show_bar_values(
    remove_y_axis: bool = True, fontsize: int = 12, position: str = "below", fmt: str = "{:.0f}", ax: mpl.axes.Axes = None
):
    if ax is None:
        ax = plt.gca()

    if remove_y_axis:
        plt.tick_params(
            axis="y",
            which="both",
            left=False,
            right=False,
            labelleft=False,
        )
        plt.grid(False)
        plt.margins(x=0)
        ax.spines["left"].set_visible(False)
        ax.spines["bottom"].set_visible(False)

    for rect in ax.patches:
        height = rect.get_height()
        ypos = height * 0.98 if position == "below" else height * 1.01
        ax.text(
            rect.get_x() + rect.get_width() / 2,
            ypos,
            fmt.format(height),
            size=fontsize,
            weight="bold",
            ha="center",
            va="top" if position == "below" else "bottom",
            color="white" if position == "below" else layout_colors["text"],
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rootsstyle']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.5.0,<4.0.0']

setup_kwargs = {
    'name': 'rootsstyle',
    'version': '0.5.1',
    'description': 'A dataroots inspired theme for matplotlib.',
    'long_description': '[![dataroots](https://dataroots.io/maintained.svg)](https://dataroots.io/)\n[![PyPI version](https://badge.fury.io/py/rootsstyle.svg)](https://badge.fury.io/py/rootsstyle)\n[![Python Versions](https://img.shields.io/badge/python->=3.8,%20<3.11-blue.svg)](https://www.python.org/downloads/)\n[![codecov](https://codecov.io/gh/datarootsio/rootsstyle/branch/main/graph/badge.svg?token=4agmmGuhtu)](https://codecov.io/gh/datarootsio/rootsstyle)\n[![semantic-release: angular](https://img.shields.io/badge/semantic--release-angular-e10079?logo=semantic-release)](https://github.com/semantic-release/semantic-release)\n\n<div align="center">\n\n# rootsstyle\n</div>\n\nA matplotlib styling package for clean, minimal dataroots themed plots. \nWorks with any visualization tools that builds upon Matplotlib (seaborn, pandas).\n\n<img align="center" src="https://github.com/datarootsio/rootsstyle/blob/main/images/examples.png?raw=true">\n\n\n\n# Installation\n### using [pip](https://pypi.org/)\n```python\npip install rootsstyle\n```\n### using [poetry](https://python-poetry.org/)\n```python\npoetry add rootsstyle\n```\n\n# Usage\n**Examples**\n\n<a href="https://colab.research.google.com/drive/1kn7YLDR4hqI9GVzeiRH9aQTEeS7HPrw7" target="_blank" rel="noopener noreferrer"><img src="https://colab.research.google.com/assets/colab-badge.svg"></a>\n\n<table width="100%">\n<tr>\n<td width="50%"> \n\n```python\nimport rootsstyle\nimport matplotlib.pyplot as plt\n\nplt.style.use(rootsstyle.style)\ny, y2 = [3, 8, 1, 10], [8, 3, 10, 2]\nplt.plot(y, label=\'y\')\nplt.plot(y2, label=\'y2\', linestyle = \'dotted\')\nrootsstyle.ylabel(\'yvalues\')\nplt.xlabel(\'x-label\')\nrootsstyle.legend()\nplt.title(\'Example plot\')\nplt.show()\n```\n\n</td>\n<td width="50%"> \n\n```python\nimport rootsstyle\nimport matplotlib.pyplot as plt\n\nplt.style.use(rootsstyle.style)\nlanguages = [\'C\', \'C++\', \'Java\', \'Python\', \'PHP\']\nstudents = [23,17,35,29,12]\nplt.bar(languages, students)\nplt.xlabel(\'Language\')\nrootsstyle.show_bar_values()\nplt.title(\'Example barplot\')\nplt.show()\n```\n</td>\n</tr>\n<tr>\n<td width="50%"><img src="https://github.com/datarootsio/rootsstyle/blob/main/images/example_lineplot.png?raw=true"></td>\n<td width="50%"> <img src="https://github.com/datarootsio/rootsstyle/blob/main/images/example_barplot.png?raw=true"></td>\n</tr>\n</table>\n\n\n**STYLING**\n\n```python\nimport rootsstyle\nimport matplotlib.pyplot as plt\n\n# globally\nplt.style.use(rootsstyle.style)\n\n# within context manager\nwith plt.style.context(rootsstyle.style):\n    # ...\n```\n\n**FUNCTIONS**\n* Place the legend to the right of the plot.<br>For lineplots, place the legend entries right of the corresponding line.\n    ```python \n    rootsstyle.legend(handles=None, labels=None, title=None)\n    ```\n\n* Place the y-label above the y-axis and rotate it, so that it is horizontal.\n    ```python \n    rootsstyle.ylabel(ylabel: str)\n    ```\n* Show barvalues at each bar. <br>Removes the y-axis (optional).<br>Bar values can be shown just \'below\' the top of each bar, or just \'above\' each bar.\n    ```python \n    rootsstyle.show_bar_values(remove_y_axis=True, fontsize=12, position="below", fmt="{:.0f}")\n    ```\n\n \n\n**COLOR PALETTE**\n<table width="100%">\n    <tr>\n        <td width="40%" align="center">The color palettes are added to the global matplotlib color registry. You can thus easily use a palette by simply providing the name in the correct location.</td>\n        <td width="60%" align="center">\n            <img src="https://github.com/datarootsio/rootsstyle/blob/main/images/palette.png?raw=true" height="350px;">\n        </td>\n    </tr>\n</table>\n\n\n# VERSIONING\n\nA [semantic versioning](https://semver.org/) scheme is used to update the version on the commit messages. This happens automatically on any push to the main branch. Only patches, minor and major changes will generate a tag, release and publishing of the package. We stick to the default [Angular Commit Message Conventions](https://github.com/angular/angular.js/blob/master/DEVELOPERS.md#-git-commit-guidelines). To use this system, commit messages should adhere to a couple of rules:\n\n1. Commits must follow the following syntax\n\n    ```\n    <type>(<scope>): <subject>\n    <BLANK LINE>\n    <body>\n    <BLANK LINE>\n    <footer>\n    ```\n\n2. Type should be one of the following:\n    * feat: A new feature\n    * fix: A bug fix\n    * docs: Documentation only changes\n    * style: Changes that do not affect the meaning of the code (formatting, missing semi-colons, etc)\n    * refactor: A code change that neither fixes a bug nor adds a feature\n    * perf: A code change that improves performance\n    * test: Adding missing or correcting existing tests\n    * chore: Changes to the build process or auxiliary tools and libraries such as documentation generation\n\n3. Body (optional) is used to motivate the change\n4. Footer (optional) is used to link to any **issues** that the commit closes and for **breaking changes**, in which case the line should start with `BREAKING CHANGE:`.\n\n# CHANGELOG\nThe [CHANGELOG.md](https://github.com/datarootsio/rootsstyle/blob/main/CHANGELOG.md) file is automatically updated upon any new releases.\n\n# License\nThis project is licensed under the terms of the MIT license.',
    'author': 'YannouRavoet',
    'author_email': 'yannou.ravoet@dataroots.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/yannouravoet',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from matplotlib.colors import (
    ListedColormap,
    LinearSegmentedColormap,
)

layout_colors = {
    "text": "#969696",
    "edges": "#38B580",
}
colors = [
    "#38B580",
    "#445BA7",
    "#BFC6E2",
    "#48DF88",
    "#34495D",
    "#A3F6B4",
    "#212121",
    "#969696",
    "#C0C2C2",
]

palettes = {
    "dataroots-default": ListedColormap(
        name="dataroots-default",
        colors=colors,
    ),
    "dataroots-green": LinearSegmentedColormap.from_list(
        name="dataroots-green",
        colors=[
            "#C3E9D9",
            "#38B580",
            "#04441C",
        ],
        gamma=1.0,
    ),
    "dataroots-blue": LinearSegmentedColormap.from_list(
        name="dataroots-blue",
        colors=[
            "#C9CFE5",
            "#445BA7",
            "#19213E",
        ],
        gamma=1.0,
    ),
    "dataroots-blue-to-green": LinearSegmentedColormap.from_list(
        name="dataroots-blue-to-green",
        colors=[
            "#19213E",
            "#38B580",
            "#B4E3CF",
        ],
        gamma=1.5,
    ),
    "dataroots-green-to-blue": LinearSegmentedColormap.from_list(
        name="dataroots-green-to-blue",
        colors=[
            "#022108",
            "#5387C8",
            "#BFC6E2",
        ],
        gamma=1.5,
    ),
}
